# Arbitrary line with intercept and slope

# Acts like axhline or axvline
# Works for 2D plots

import matplotlib.pyplot as plt 
import numpy as np    

def axaline(slope, intercept,style=':r',label='name'):
    """Plot a line from slope and intercept"""
    axes = plt.gca()
    x_vals = np.array(axes.get_xlim())
    y_vals = intercept + slope * x_vals
    plt.plot(x_vals, y_vals, style, label=label)



# 95% Confidence Inerval Ellipse on PCA scores


import numpy as np
from scipy import stats


def PCA_ellipse(T):

    N = T.shape[0]
    nbPC = T.shape[1]
    Fdf1 = nbPC
    Fdf2 = N-nbPC
    T2_95 = nbPC*(N-1)/(N-nbPC)*stats.f.ppf(0.95,Fdf1,Fdf2)
    
    u = 0     #x-position of the center
    v = 0    #y-position of the center
    a = np.sqrt(np.var(T[:,0]) * T2_95)      #radius on the x-axis
    b = np.sqrt(np.var(T[:,1]) * T2_95)    #radius on the y-axis
    
    t = np.linspace(0, 2*np.pi, 100)
    H = u+a*np.cos(t)
    V = v+b*np.sin(t)
    
    H = H[np.newaxis].T
    V = V[np.newaxis].T
    
    Tellipse = np.concatenate((H,V),axis=1)
    
    return Tellipse



# ICA: independent component analysis

import numpy as np

def ICA(X):
    
    # X must be centered

    [D, E] = np.linalg.eig(X.T@X)
    D = D*np.eye(len(D))
    Dinv = np.linalg.pinv(D)
    P = E@np.sqrt(Dinv)@E.T
    T = X@P 
    
    return T,P


# cPCA: consensus PCA

import numpy as np

def cPCA(nbPC, *args):
    ''' 
    
    cPCA: consensus PCA
    [Tt, T_all, W, P_all, SSX, SSX_all] = cPCA(nbPC, X1, X2, Xi, ...)
    
    INPUT
    nbPC [1 x 1] <numpy.ndarray>
        number of components in the PCA decomposition
    Xi [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    
    Tt [n x nbPC] <numpy.ndarray>
        Super scores
    T_all [nbPC] <list>
        Scores for each block (corresponding to Xi)
    W [m x nbPC] <numpy.ndarray>
        Weights of each block (ie block loadings)
        m blocks
    P_all [nbPC] <list>
        Loadings for each block (corresponding to Xi)
    SSX [nbPC x 1] <numpy.ndarray>
        Sum of squared variance explained by each component
    SSX_all [nbPC x m] <list>
        Sum of squared variance explained by each component for each block
        m blocks
        
    '''
     
#%% Load data
    # Tuple containing n block matrices
    varargin = args
    n_block = len(varargin)
    
    # Creation of matrix X containing n blocks  
    X = []
    X_all = []
    for i in range(n_block):
        xi = varargin[i]
        if (i==0):
            X = xi
        else:
            X = np.append(X,xi,axis=1)

    
    n,k = X.shape

    # Save copy
    varargin0 = varargin
    X0 = X
    
    #%% Initialization 
    T = np.zeros((n, n_block))
    Tt = np.zeros((n,nbPC))
    W = np.zeros((n_block,nbPC))

    X_all = []

    # For creation of tuple containing T for each block
    T_all = []
    T_all_temp = []
    for i in range(n_block):
        Ti = np.zeros((n,nbPC))
        T_all.append(Ti)
    
    # Creation of tuple containing P for each block
    P_all = []
    P_all_temp = []
    for i in range(n_block): 
        Xi = varargin[i]
        n, ki = Xi.shape
        Pi = np.zeros((ki, nbPC))
        P_all.append(Pi)

    # Sum of squares
    SSXi = np.zeros((nbPC,n_block)) 
    
    #%% PCA
    
    for a in range(nbPC):
        tt = np.ones((n,1))
        tt_temp = np.ones((n,1))
        error_tt = 1
    
        while (error_tt > 1E-20):
        
            for i in range(n_block):
                Xi = varargin[i]
                pi = (Xi.T@tt)/(tt.T@tt)
                pi = pi/np.linalg.norm(pi)
                ti = Xi@pi
            
                T[:,i] = np.squeeze(ti)
        
            wt = (T.T@tt)/(tt.T@tt)
            wt = wt/np.linalg.norm(wt)
            tt = T@wt
        
            error_tt = sum(np.power((tt - tt_temp),2))
            tt_temp = tt
        
        for i in range(n_block):
            # Deflation 
            Xi = varargin[i]
            pi = (Xi.T@tt)/(tt.T@tt) 
            Xi = Xi - tt@pi.T
            X_all.append(Xi)
                         
            # Sum of squares of X
            Xi_hat = tt@pi.T
            Xi_0 = varargin0[i]
            ssXi_0 = np.sum(np.sum(Xi_0*Xi_0))
            ssXi = np.sum(np.sum(Xi_hat*Xi_hat))
            ssXi = ssXi/ssXi_0
            
            SSXi[a,i] = ssXi
            
            # Save pi 
            Pi = P_all[i]
            Pi[:,a] = np.squeeze(pi)
            P_all_temp.append(Pi)
        P_all = P_all_temp
        P_all_temp = []
        
    
        # Update of deflated blocks
        varargin = X_all
        X_all = []                         
    
        # Save
        Tt[:,a] = np.squeeze(tt)
        W[:,a] = np.squeeze(wt)
    
        for i in range(n_block):
            Ti = T_all[i]
            Ti[:,a] = T[:,i]
            T_all_temp.append(Ti)
        T_all = T_all_temp
        T_all_temp = []  
    
    SSX = np.sum(SSXi,axis=1)/n_block  
    
    SSX_all = SSXi
    
    return Tt, T_all, W, P_all, SSX, SSX_all
